import re

from types import MethodType

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models.functions import Cast
from django.utils.translation import ugettext_lazy as _

from chamber.models import SmartModel, SmartQuerySet


def camel_to_snake(name):
  name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
  return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()


def add_objs(self, *objects):
    for obj in objects:
        self.get_or_create(
            object_ct=ContentType.objects.get_for_model(obj),
            object_id=obj.pk
        )


def clear_objs(self):
    self.all().delete()


def set_objs(self, *objects):
    self.clear()
    for obj in objects:
        self.create(
            object_ct=ContentType.objects.get_for_model(obj),
            object_id=obj.pk
        )


def remove_objs(self, *objects):
    for obj in objects:
        self.filter(
            object_ct=ContentType.objects.get_for_model(obj),
            object_id=obj.pk
        ).delete()


class RelatedObjectQuerySet(SmartQuerySet):

    def annotate_object_pks(self, model_class):
        pk_field = model_class._meta.pk
        if isinstance(pk_field, models.AutoField):
            pk_field = models.IntegerField()
        return self.filter(
            object_ct=ContentType.objects.get_for_model(model_class)
        ).annotate(
            object_pk=Cast('object_id', output_field=pk_field)
        )

    def get_objects(self, model_class):
        return model_class.objects.filter(pk__in=self.annotate_object_pks(model_class).values('object_pk'))

    def get_object_pks(self, model_class):
        return self.annotate_object_pks(model_class).values_list('object_pk', flat=True)


class GenericManyToManyManager(models.Manager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self._is_related_manger():
            self.add = MethodType(add_objs, self)
            self.set = MethodType(set_objs, self)
            self.clear = MethodType(clear_objs, self)
            self.remove = MethodType(remove_objs, self)

    def _is_related_manger(self):
        return self.__class__.__module__ == 'django.db.models.fields.related_descriptors'


class GenericManyToMany(SmartModel):

    object_ct = models.ForeignKey(
        verbose_name=_('content type of the related object'),
        to=ContentType,
        null=False,
        blank=False,
        on_delete=models.CASCADE
    )
    object_id = models.TextField(
        verbose_name=_('ID of the related object'),
        null=False,
        blank=False,
        db_index=True
    )
    object = GenericForeignKey(
        'object_ct',
        'object_id'
    )
    object.verbose_name = _('related object')

    objects = GenericManyToManyManager.from_queryset(RelatedObjectQuerySet)()

    class Meta:
        abstract = True
        ordering = ('-created_at',)


def create_generic_many_to_many_intermediary_model(field, klass):
    from_ = camel_to_snake(klass.__name__).lower()

    name = '{}GenericManyToManyRelation'.format(klass._meta.object_name)
    meta = type('Meta', (), {
        'app_label': klass._meta.app_label,
        'unique_together': (from_, 'object_ct', 'object_id'),
        'apps': field.model._meta.apps,
    })
    return type(name, (GenericManyToMany,), {
        'Meta': meta,
        '__module__': klass.__module__,
        from_: models.ForeignKey(
            klass,
            on_delete=models.CASCADE,
            related_name=field.name
        ),
    })


class GenericManyToManyField:

    def __init__(self, through=None):
        self.through = through

    def contribute_to_class(self, cls, name, **kwargs):
        self.model = cls
        self.name = name
        self.through = self.through or create_generic_many_to_many_intermediary_model(self, cls)



from django.db import models

from generic_m2m_field.models import GenericManyToManyField


class GenericManyToManyModel(models.Model):

    related_objects = GenericManyToManyField()


class OneRelatedObject(models.Model):
    pass


class SecondRelatedObject(models.Model):

    id = models.CharField(primary_key=True, max_length=10)



from germanium.test_cases.default import GermaniumTestCase
from germanium.tools import assert_equal

from apps.app.models import GenericManyToManyModel, OneRelatedObject, SecondRelatedObject


class GenericManyToManyTestCase(GermaniumTestCase):

    def test_generic_m2m_should_add_related_object(self):
        m2m_inst = GenericManyToManyModel.objects.create()
        related_object_inst1 = OneRelatedObject.objects.create()
        related_object_inst2 = SecondRelatedObject.objects.create(id='unique')
        related_object_inst3 = OneRelatedObject.objects.create()

        assert_equal(m2m_inst.related_objects.count(), 0)
        m2m_inst.related_objects.add(related_object_inst1)
        assert_equal(m2m_inst.related_objects.count(), 1)

        m2m_inst.related_objects.add(related_object_inst2, related_object_inst3)

        assert_equal(m2m_inst.related_objects.count(), 3)

        # Duplicate objects are not added
        m2m_inst.related_objects.add(related_object_inst2)
        assert_equal(m2m_inst.related_objects.count(), 3)

    def test_generic_m2m_should_clear_related_object(self):
        m2m_inst = GenericManyToManyModel.objects.create()
        related_object_inst1 = OneRelatedObject.objects.create()
        related_object_inst2 = SecondRelatedObject.objects.create(id='unique')

        m2m_inst.related_objects.add(related_object_inst1, related_object_inst2)
        assert_equal(m2m_inst.related_objects.count(), 2)

        m2m_inst.related_objects.clear()
        assert_equal(m2m_inst.related_objects.count(), 0)

    def test_generic_m2m_should_set_related_object(self):
        m2m_inst = GenericManyToManyModel.objects.create()
        related_object_inst1 = OneRelatedObject.objects.create()
        related_object_inst2 = SecondRelatedObject.objects.create(id='unique')
        related_object_inst3 = OneRelatedObject.objects.create()

        m2m_inst.related_objects.add(related_object_inst1, related_object_inst2)
        assert_equal(m2m_inst.related_objects.count(), 2)

        m2m_inst.related_objects.set(related_object_inst3)
        assert_equal(m2m_inst.related_objects.count(), 1)
        assert_equal(m2m_inst.related_objects.get().object_id, str(related_object_inst3.pk))

    def test_generic_m2m_should_remove_related_object(self):
        m2m_inst = GenericManyToManyModel.objects.create()
        related_object_inst1 = OneRelatedObject.objects.create()
        related_object_inst2 = SecondRelatedObject.objects.create(id='unique')
        related_object_inst3 = OneRelatedObject.objects.create()

        m2m_inst.related_objects.add(related_object_inst1, related_object_inst2, related_object_inst3)
        assert_equal(m2m_inst.related_objects.count(), 3)

        m2m_inst.related_objects.remove(related_object_inst2)
        assert_equal(m2m_inst.related_objects.count(), 2)
        assert_equal(
            set(m2m_inst.related_objects.values_list('object_id', flat=True)),
            {str(related_object_inst1.pk), str(related_object_inst3.pk)}
        )

    def test_generic_m2m_manager_should_return_object_pks_according_to_model_class(self):
        m2m_inst = GenericManyToManyModel.objects.create()
        related_object_inst1 = OneRelatedObject.objects.create()
        related_object_inst2 = SecondRelatedObject.objects.create(id='unique')
        related_object_inst3 = OneRelatedObject.objects.create()

        m2m_inst.related_objects.add(related_object_inst1, related_object_inst2, related_object_inst3)

        assert_equal(list(m2m_inst.related_objects.get_object_pks(SecondRelatedObject)), [related_object_inst2.pk])
        assert_equal(
            list(m2m_inst.related_objects.get_object_pks(OneRelatedObject)),
            [related_object_inst1.pk, related_object_inst3.pk]
        )

    def test_generic_m2m_manager_should_return_object_according_to_model_class(self):
        m2m_inst = GenericManyToManyModel.objects.create()
        related_object_inst1 = OneRelatedObject.objects.create()
        related_object_inst2 = SecondRelatedObject.objects.create(id='unique')
        related_object_inst3 = OneRelatedObject.objects.create()

        m2m_inst.related_objects.add(related_object_inst1, related_object_inst2, related_object_inst3)

        assert_equal(list(m2m_inst.related_objects.get_objects(SecondRelatedObject)), [related_object_inst2])
        assert_equal(
            list(m2m_inst.related_objects.get_objects(OneRelatedObject)),
            [related_object_inst1, related_object_inst3]
        )



# Generated by Django 2.2 on 2020-02-03 20:26

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='GenericManyToManyModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
        ),
        migrations.CreateModel(
            name='OneRelatedObject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
        ),
        migrations.CreateModel(
            name='SecondRelatedObject',
            fields=[
                ('id', models.CharField(max_length=10, primary_key=True, serialize=False)),
            ],
        ),
        migrations.CreateModel(
            name='GenericManyToManyModelGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_id', models.TextField(db_index=True, verbose_name='ID of the related object')),
                ('generic_many_to_many_model', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='related_objects', to='app.GenericManyToManyModel')),
                ('object_ct', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType', verbose_name='content type of the related object')),
            ],
            options={
                'unique_together': {('generic_many_to_many_model', 'object_ct', 'object_id')},
            },
        ),
    ]



from setuptools import setup, find_packages


setup(
    name='django-generic-m2m-field',
    version='0.0.1',
    description="Django generic many to many field library.",
    keywords='django, fields, generic',
    author='Lubos Matl',
    author_email='matllubos@gmail.com',
    url='https://github.com/druids/django-generic-m2m-field',
    license='MIT',
    package_dir={'generic_m2m_field': 'generic_m2m_field'},
    include_package_data=True,
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: Czech',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: Site Management',
    ],
    install_requires=[
        'django>=1.11',
        'django-chamber>=0.5.3',
    ],
    zip_safe=False
)
